/*	Application		IPC receiver-side using socket (UDP)		
	File			IPC_socket_Receiver_with_Reply.java
	Purpose			Demonstration of local IPC, using a blocking socket
	Author			Richard John Anthony
	Date			January 2015
	
	Function		1. Creates a socket, and binds to local address (implcitly uses INADDR_ANY)
				2. Waits for a message to arrive on its port (8007)
				3. Displays the message 
                                4. Reverses the message contents
                                5. Sends the modified message back as a reply to the sender (to port 8008)
                                6. Exits once a key is pressed

    To run the project from the command line, go to the dist folder and type: java -jar "IPC_socket_Receiver_with_Reply.jar"
*/
package ipc_socket_receiver_with_reply;

import java.io.Console;
import java.net.*;

public class IPC_socket_Receiver_with_Reply 
{
    private static  int BUFFER_SIZE = 100;
    private static  int RECEIVE_PORT = 8007;
    private static  int REPLY_PORT = 8008;  // The port to send the reply to (the port the original sender will be listening on)
    private static DatagramSocket m_IPC_SOCKET;
    private static InetAddress m_SenderIPAddress;   // Will hold the address and port of the sender after a message has been received
    static Console con;
    
    public static void main(String[] args)
    {
        con = System.console();
        if (null == con)
        {
            System.out.println("!! Error accessing console: Run in a command window !!");
            System.exit(-1);
        }
        System.out.println("Local Inter-Process Communication (receiver side).");
        System.out.println("This program waits for a message to arrive on port 8007 (decimal).");
        System.out.println("The message is displayed, and the message contents are reversed.");
        System.out.println("The modified message is sent back to the sender as a reply, on port 8008.");
        System.out.println("The process exits when a key is pressed.");
        System.out.println("-------------------------------------------------------------------------");
        System.out.println("Waiting for message ...   ");

        IPC_socket_Receiver_with_Reply Socket_Receiver = new IPC_socket_Receiver_with_Reply();
    }

    IPC_socket_Receiver_with_Reply()
    {
        CreateSocket();
        
        // Wait for / receive the message
        String sMessageString = ReadMessageFromReceiveBufferOrWait();
        System.out.printf("   Message received:       %s\n", sMessageString);
        System.out.printf("   from sender at address: %s\n", m_SenderIPAddress.toString());

        // Reverse the received message contents in the buffer
        sMessageString = ReverseString(sMessageString);
        System.out.printf("   Reversed reply message: %s\n", sMessageString);

        // Send the reply message to the original message sender
        SendMessage(sMessageString);
        System.out.println("   Reply message sent");

        CloseSocketAndExit();
    }

    void CreateSocket()
    {
        try
        {
            m_IPC_SOCKET = new DatagramSocket(RECEIVE_PORT); // Create socket and bind it to the receive port (uses the local address)
            m_IPC_SOCKET.setSoTimeout(0); // A timeout value of 0 is interpreted as infinity i.e blocking-IO behaviour
        }
        catch (SocketException ex)  // Catch exceptions from the socket operations
        {
            System.out.printf("Error %s\n", ex.toString());
            CloseSocketAndExit();
        }
    }

    String ReadMessageFromReceiveBufferOrWait()
    {
        String sMessage = "";
        try
        {
            byte[] bRecvBuf = new byte[BUFFER_SIZE];
            DatagramPacket RecvPacket = new DatagramPacket(bRecvBuf, BUFFER_SIZE);
            m_IPC_SOCKET.receive(RecvPacket);

            // Place the sender's address into the m_SenderIPAddress
            m_SenderIPAddress = RecvPacket.getAddress();
            sMessage = new String(RecvPacket.getData(), 0, RecvPacket.getLength());  
        }
        catch(Exception ex)
        {
            System.out.printf("Receive failed: %s\n", ex.toString());
            CloseSocketAndExit();
        }
        return sMessage;
    }
       
    String ReverseString(String sStr)
    {
        return new StringBuffer(sStr).reverse().toString();
    }

    void SendMessage(String sSendString)
    {
        byte[] bSendBuf = new byte[BUFFER_SIZE];
        bSendBuf = sSendString.getBytes();
        try
        {
            DatagramPacket SendPacket = new DatagramPacket(bSendBuf, sSendString.length(),
                                    m_SenderIPAddress /*The address to send to*/, REPLY_PORT); 
            m_IPC_SOCKET.send(SendPacket);
        }
        catch(Exception ex)
        {
            System.out.printf("Send failed: %s\n", ex.toString());
            CloseSocketAndExit();
        }
    }

    void CloseSocketAndExit()
    {
        try
        {
            m_IPC_SOCKET.close();
            System.out.println("Reply socket closed");
        }
        catch(Exception ex)
        {
        }
        System.out.println("(press a key to continue)");
        con.readLine();	// Wait for user to type a key before the program exits
        System.exit(1);
    }
}