/*	Application		IPC sender-side using socket (UDP)		
	File			IPC_socket_Sender_with_Reply.java
	Purpose			Demonstration of local IPC, using a blocking socket
	Author			Richard John Anthony
	Date			January 2015
	
	Function		1. Creates a socket
                                2. Waits for user to type a message
                                3. Sends the message to the the application port (8007) on the local computer (uses local loopback address 127.0.0.1)
                                4. Waits for a reply message to be returned (on port 8008)
                                5. Displays the reply message
                                6. Exits once the reply message has been displayed

    To run the project from the command line, go to the dist folder and type: java -jar "IPC_socket_Sender_with_Reply.jar"
*/
package ipc_socket_sender_with_reply;

import java.io.Console;
import java.net.*;

public class IPC_socket_Sender_with_Reply
{
    private static int BUFFER_SIZE = 100;
    private static int SEND_PORT = 8007; // The port to send to
    // Use a different port for binding to, so both sender and receiver processes can run on the same computer
    private static int ReceiveReply_PORT = 8008; 
    private static DatagramSocket m_IPC_SOCKET;
    private static InetAddress m_SenderIPAddress;   // Will hold the address and port of the sender after a message has been received
    static Console con;
    
    public static void main(String[] args)
    {
        con = System.console();
        if (null == con)
        {
            System.out.println("!! Error accessing console: Run in a command window !!");
            System.exit(-1);
        }

        System.out.println("Local Inter-Process Communication (sender side).");
        System.out.println("This program waits for the user to type a message,");
        System.out.println("and sends it to port 8007 on the local computer, and");
        System.out.println("waits for a reply message to be sent back on port 8008.");
        System.out.println("--------------------------------------------------");
        System.out.println("Type a  message to send ...   ");

        IPC_socket_Sender_with_Reply Socket_Sender = new IPC_socket_Sender_with_Reply();
    }

    IPC_socket_Sender_with_Reply()
    {
        CreateSocket();

        // Wait for the user to type a message
        String sMessageString = con.readLine(); // Get a string which may include spaces();
        
        // Send the message that the user typed, to the sepcified port on the local computer
        SendMessage(sMessageString);
        System.out.printf("   Message \"%s\" sent\n", sMessageString);
        System.out.println("   Waiting for reply ...");

	// Message has been sent - now wait for the reply
	sMessageString = ReadMessageFromReceiveBufferOrWait();
        System.out.printf("   Message received:       %s\n", sMessageString);
        System.out.printf("   from sender at address: %s\n", m_SenderIPAddress.toString());

        CloseSocketAndExit();
    }

    void CreateSocket()
    {
        try
        {
            m_IPC_SOCKET = new DatagramSocket(ReceiveReply_PORT); // Create socket and bind it a specified port (uses the local address)
            m_IPC_SOCKET.setSoTimeout(1000); // Non-blocking with a timeout value of 1000 milliseconds
        }
        catch (SocketException ex)  // Catch exceptions from the socket operations
        {
            System.out.printf("Error %s\n", ex.toString());
            CloseSocketAndExit();
        }
    }
    
    void SendMessage(String sSendString)
    {
        byte[] bSendBuf = new byte[BUFFER_SIZE];
        bSendBuf = sSendString.getBytes();
        try
        {
            InetAddress SendIPAddress = InetAddress.getLoopbackAddress(); // Assume receiver is at same address as sender
            DatagramPacket SendPacket = new DatagramPacket(bSendBuf, sSendString.length(),
                                    SendIPAddress /*The address to send to*/, SEND_PORT); 
            m_IPC_SOCKET.send(SendPacket);
        }
        catch(Exception ex)
        {
            System.out.printf("Send failed: %s\n", ex.toString());
            CloseSocketAndExit();
        }
    }
    
    String ReadMessageFromReceiveBufferOrWait()
    {
        String sMessage = "";
        try
        {
            byte[] bRecvBuf = new byte[BUFFER_SIZE];
            DatagramPacket RecvPacket = new DatagramPacket(bRecvBuf, BUFFER_SIZE);
            m_IPC_SOCKET.receive(RecvPacket);

            // Place the sender's address into the m_SenderIPAddress
            m_SenderIPAddress = RecvPacket.getAddress();
            sMessage = new String(RecvPacket.getData(), 0, RecvPacket.getLength());  
        }
        catch(Exception ex)
        {
            System.out.printf("Receive failed: %s\n", ex.toString());
            CloseSocketAndExit();
        }
        return sMessage;
    }

    void CloseSocketAndExit()
    {
        try
        {
            m_IPC_SOCKET.close();
            System.out.println("Socket closed");
        }
        catch(Exception ex)
        {
        }
        System.out.println("(press a key to continue)");
        con.readLine();	// Wait for user to type a key before the program exits
        System.exit(1);
    }
}