/*	Application		Socket-based IPC threads example	
	File			Multithreaded_IPC.java
	Purpose			Demonstration of local IPC, and multithreading using a blocking socket
	Author			Richard John Anthony
	Date			February 2015
	
	Usage			Requires two command-line arguments {send port, recv port}

	Function		1. Creates a pair of sockets (one for receiving, one for sending)
                                2. Binds to the user-provided receive-port number
                                3. Starts a worker thread to handle receive activity
                                4. Handles send activity on the main thread
                                5. Exits once the string "Quit" (or "QUIT", "quit") is entered

    To run the project from the command line, go to the dist folder and type: java -jar "Multithreaded_IPC.jar"
*/

package Multithreaded_IPC;

import java.io.Console;
import java.net.*;

public class Multithreaded_IPC implements Runnable
{
    private static  int BUFFER_SIZE = 100;
    private static DatagramSocket m_IPC_SendSOCKET;
    private static DatagramSocket m_IPC_RecvSOCKET;
    Thread m_RecvThread;
    private static Boolean m_bMainThreadRunning;   
    static Console con;
    
    public static void main(String[] args)
    {
        int iRecvPort = 0;
        int iSendPort = 0;
    
        con = System.console();
        if (null == con)
        {
            System.out.println("!! Error accessing console: Run in a command window !!");
            System.exit(-1);
        }
        
        if (2 != args.length)
        {   // User did not provide the required parameters

            // Note that in C++ the executable name is provided as command line argument [0]. 
            // In Java only the user-provided arguments are passed to the process in the args array,
            // (i.e. NOT including the executable name), so a different technique is necessary to get the program name.
            String sProgramName = new java.io.File(Multithreaded_IPC.class.getProtectionDomain()
                                                .getCodeSource().getLocation().getPath()).getName();
            System.out.printf("Usage: %s SendPort_number ReceivePort_number\n", sProgramName);
            CloseSocketsAndExit(true);
        }
        try 
        {
            iSendPort = Integer.parseInt(args[0]);
            if(1023 >= iSendPort)
            {   // Send Port value is in reserved port range
                System.out.println("Send port number must be 1024 or greater (up to 65535)");
                CloseSocketsAndExit(true);
            }
            iRecvPort = Integer.parseInt(args[1]);
            if(1023 >= iRecvPort)
            {   // Receive port value is in reserved port range
                System.out.println("Receive port number must be 1024 or greater (up to 65535)");
                CloseSocketsAndExit(true);
            }
        }
        catch (NumberFormatException Ex)
        {
            System.out.println("Both port arguments must be integers");
            System.exit(1);
        }

        System.out.println("Multithreaded Local Inter-Process Communication");
        System.out.printf("This process sends messages to port %d and receives messages on port %d\n", iSendPort, iRecvPort);
        System.out.println("Sending is performed on main thread, receiving is performed by a worker thread");
        System.out.println("------------------------------------------------------------------------------");

        CreateSendSocket();
        CreateRecvSocket(iRecvPort);

	// Create and start worker thread (performs Receive in loop)
 	Multithreaded_IPC ReceiveThread = new Multithreaded_IPC();
	ReceiveThread.Start(); // The ReceiveThread runs asynchronously with the main() thread

        // The Send thread simply continues from here (it is the original 'main' thread of execution)
	m_bMainThreadRunning = true;
        String sIn = "";
        while(true)
        {
            System.out.println("SendThread: Type a message to send (\"Quit\" to quit): ");
            try
            {
                sIn = con.readLine();
                if(sIn.length() >= 4 && sIn.substring(0, 4).toUpperCase().equals("QUIT"))
                {
                    m_bMainThreadRunning = false;   // Signal to Receive Thread to stop its operation
                    System.out.println("Quiting");
                    CloseSocketsAndExit(false);
                }
            }
            catch (Exception Ex)
            {   // Control-C for example causes us to come here
                m_bMainThreadRunning = false;   // Signal to Receive Thread to stop its operation
                System.out.println("Exception occurred, Quiting");
                CloseSocketsAndExit(false);
            }
            int iNumberOfBytesSent = SendMessage(iSendPort, sIn);
            if(-1 != iNumberOfBytesSent)
            {
                System.out.printf("SendThread  Sent: %s Number of bytes sent: %d\n", sIn, iNumberOfBytesSent);
            }
        }
    }

    static int SendMessage(int iSendPort, String sSendString)
    {
        byte[] bSendBuf = new byte[BUFFER_SIZE];
        bSendBuf = sSendString.getBytes();
        int iNumBytesSent = 0;
        try
        {
            InetAddress IPAddress = InetAddress.getLocalHost();
            // Uncomment the statement below for diagnostic purposes
            // System.out.printf("Send assumes local receiver, uses local address: %s:%d\n", 
            //                                        IPAddress.toString(), iSendPort);
            DatagramPacket SendPacket = new DatagramPacket(bSendBuf, sSendString.length(),
                    IPAddress /*uses local address as the address to send to*/, iSendPort); 
            m_IPC_SendSOCKET.send(SendPacket);
            iNumBytesSent = SendPacket.getLength();  
        }
        catch(Exception Ex)
        {
            System.out.printf("Send failed: %s\n", Ex.toString());
            return -1; // Signal that an error occurred
        }
        return iNumBytesSent;
    }

    static void CreateSendSocket()
    {
        try
        {
            m_IPC_SendSOCKET = new DatagramSocket(); // Bind to any available port (uses the local address)
            m_IPC_SendSOCKET.setSoTimeout(100); // A timeout value of 100ms (i.e. the send socket is non-blocking)
        }
        catch (SocketException Ex)  // Catch exceptions from the socket operations
        {
            System.out.printf("Error %s\n", Ex.toString());
            CloseSocketsAndExit(true);
        }
    }

    static void CreateRecvSocket(int iPort)
    {
        try
        {
            m_IPC_RecvSOCKET = new DatagramSocket(iPort); // Create socket and bind it to the receive port (uses the local address)
            m_IPC_RecvSOCKET.setSoTimeout(0); // A timeout value of 0 is interpreted as infinity i.e blocking-IO behaviour
        }
        catch (SocketException Ex)  // Catch exceptions from the socket operations
        {
            System.out.printf("Error %s\n", Ex.toString());
            CloseSocketsAndExit(true);
        }
    }
    
    static void CloseSocketsAndExit(Boolean bWaitForKey)
    {
        try
        {
            m_IPC_SendSOCKET.close();
            System.out.println("Send socket closed");
        }
        catch(Exception Ex)
        {
        }
        try
        {
            m_IPC_RecvSOCKET.close();
            System.out.println("Receive socket closed");
        }
        catch(Exception Ex)
        {
        }
        if(true == bWaitForKey)
        {
            System.out.println("(press a key to continue)");
            con.readLine();	// Wait for user to type a key before the program exits
        }
        System.exit(1);
    }

    // ******************************* Beginning of Receive Thread code *******************************
    Multithreaded_IPC()
    {
        m_RecvThread = new Thread(this); // Create the receiver thread
    }

    public void Start()
    {   // The means of externally calling the thread's start() method (which invokes run()).
        m_RecvThread.start();
    }

    public void run()
    {
        while (true == Multithreaded_IPC.m_bMainThreadRunning)
        {
            ReadMessageFromReceiveBufferOrWait();
        }
    }

    String ReadMessageFromReceiveBufferOrWait()
    {
        String sReceivedStr = "";
        if(true == m_bMainThreadRunning)
        {
            System.out.println("ReceiveThread: Waiting for message ...   ");
            try
            {
                byte[] bRecvBuf = new byte[BUFFER_SIZE];
                DatagramPacket RecvPacket = new DatagramPacket(bRecvBuf, BUFFER_SIZE);
                m_IPC_RecvSOCKET.receive(RecvPacket);
                sReceivedStr = new String(RecvPacket.getData(), 0, RecvPacket.getLength());  
                System.out.printf("RecvThread  Received: %s   number of bytes received: %d\n", sReceivedStr, sReceivedStr.length());
            }
            catch(Exception Ex)
            {
                if(true == m_bMainThreadRunning) // Test again, as the flag may have become set whilst this thread was blocked
                {
                    System.out.printf("Receive failed: %s\n", Ex.toString());
                    CloseSocketsAndExit(true);
                }
            }
        }
        return sReceivedStr;
    }
    // ********************************* End of Receive Thread code *********************************
}