/*	Application		Socket-based IPC threads example	
	File			Multithreaded_IPC_two_worker_threads.java
	Purpose			Demonstration of local IPC, and multithreading using a blocking socket
	Author			Richard John Anthony
	Date			February 2015
	
	Usage			Requires two command-line arguments {send port, recv port}

	Function		1. Creates a pair of sockets (one for receiving, one for sending)
                                2. Binds to the user-provided receive-port number
                                3. Starts a worker thread to handle receive activity
                                4. Starts a worker thread to handle send activity
                                4. Main thread is reserved for application business logic (null in this simple demonstration)
                                5. Exits once the string "Quit" (or "QUIT", "quit") is entered

    To run the project from the command line, go to the dist folder and type: java -jar "Multithreaded_IPC_two_worker_threads.jar"
*/

package Multithreaded_IPC_two_worker_threads;

import java.io.Console;
import java.net.*;

public class Multithreaded_IPC_two_worker_threads
{
    public static  int BUFFER_SIZE = 100;
    public static DatagramSocket m_IPC_SendSOCKET;
    public static DatagramSocket m_IPC_RecvSOCKET;
    public static int m_iSendPort = 0;  // Overwritten by command-line provided value
    
    // Dedicated flags are used, one to run/stop each thread, to prevent access races
    public static Boolean m_bMainThreadRunning;   
    
    private static Receive_Thread m_ReceiveThread;
    private static Send_Thread m_SendThread;
    
    public static Console con;
    
    public static void main(String[] args)
    {
        int iRecvPort = 0;
    
        con = System.console();
        if (null == con)
        {
            System.out.println("!! Error accessing console: Run in a command window !!");
            System.exit(-1);
        }
        
        if (2 != args.length)
        {   // User did not provide the required parameters

            // Note that in C++ the executable name is provided as command line argument [0]. 
            // In Java only the user-provided arguments are passed to the process in the args array,
            // (i.e. NOT including the executable name), so a different technique is necessary to get the program name.
            String sProgramName = new java.io.File(Multithreaded_IPC_two_worker_threads.class.getProtectionDomain()
                                                .getCodeSource().getLocation().getPath()).getName();
            System.out.printf("Usage: %s SendPort_number ReceivePort_number\n", sProgramName);
            CloseSocketsAndExit(true);
        }
        try 
        {
            m_iSendPort = Integer.parseInt(args[0]);
            if(1023 >= m_iSendPort)
            {   // Send Port value is in reserved port range
                System.out.println("Send port number must be 1024 or greater (up to 65535)");
                CloseSocketsAndExit(true);
            }
            iRecvPort = Integer.parseInt(args[1]);
            if(1023 >= iRecvPort)
            {   // Receive port value is in reserved port range
                System.out.println("Receive port number must be 1024 or greater (up to 65535)");
                CloseSocketsAndExit(true);
            }
        }
        catch (NumberFormatException Ex)
        {
            System.out.println("Both port arguments must be integers");
            System.exit(1);
        }

        System.out.println("Multithreaded Local Inter-Process Communication");
        System.out.printf("This process sends messages to port %d and receives messages on port %d\n", m_iSendPort, iRecvPort);
        System.out.println("Sending and receiving is performed by a pair of dedicated worker threads.");
        System.out.println("The main thread is reserved for application business logic (none in this demo).");
        System.out.println("-------------------------------------------------------------------------------");

        CreateSendSocket();
        CreateRecvSocket(iRecvPort);

	// Create and start receive worker thread
 	m_ReceiveThread = new Receive_Thread();
	m_ReceiveThread.Start(); // The ReceiveThread runs asynchronously with the main() thread

        // Create and start send worker thread
 	m_SendThread = new Send_Thread();
	m_SendThread.Start(); // The ReceiveThread runs asynchronously with the main() thread

        // The main thread continues from here (it is the original 'main' thread of execution)
        m_bMainThreadRunning = true;
        m_ReceiveThread.m_bRecvThread_Running = true;
        m_SendThread.m_bSendThread_Running = true;
    
        while (true)
        {   // This loop is the main thread, which continues to run asynchronously with the two worker threads
            if (true == m_bMainThreadRunning)
            {
                System.out.println(" -MainThread- ");
                try
                {
                    Thread.sleep(5000);	// 5 second delay
                }
                catch (InterruptedException Ex)
                {
                }
            }
        }
    }
        
    static void CreateSendSocket()
    {
        try
        {
            m_IPC_SendSOCKET = new DatagramSocket(); // Bind to any available port (uses the local address)
            m_IPC_SendSOCKET.setSoTimeout(100); // A timeout value of 100ms (i.e. the send socket is non-blocking)
        }
        catch (SocketException Ex)  // Catch exceptions from the socket operations
        {
            System.out.printf("Error %s\n", Ex.toString());
            CloseSocketsAndExit(true);
        }
    }

    static void CreateRecvSocket(int iPort)
    {
        try
        {
            m_IPC_RecvSOCKET = new DatagramSocket(iPort); // Create socket and bind it to the receive port (uses the local address)
            m_IPC_RecvSOCKET.setSoTimeout(0); // A timeout value of 0 is interpreted as infinity i.e blocking-IO behaviour
        }
        catch (SocketException Ex)  // Catch exceptions from the socket operations
        {
            System.out.printf("Error %s\n", Ex.toString());
            CloseSocketsAndExit(true);
        }
    }
    
    static void CloseSocketsAndExit(Boolean bWaitForKey)
    {
        // Signal to all Threads to stop their operation
        m_bMainThreadRunning = false;
        m_ReceiveThread.m_bRecvThread_Running = false;
        m_SendThread.m_bSendThread_Running = false;
  
        try
        {
            m_IPC_SendSOCKET.close();
            System.out.println("Send socket closed");
        }
        catch(Exception Ex)
        {
        }
        try
        {
            m_IPC_RecvSOCKET.close();
            System.out.println("Receive socket closed");
        }
        catch(Exception Ex)
        {
        }
        if(true == bWaitForKey)
        {
            System.out.println("(press a key to continue)");
            con.readLine();	// Wait for user to type a key before the program exits
        }
        System.exit(1);
    }
}